
#include "../shim.h"
#include "event.fdh"

#ifndef GET_X_LPARAM
	#define GET_X_LPARAM(lp) ((int)(short)LOWORD(lp))
	#define GET_Y_LPARAM(lp) ((int)(short)HIWORD(lp))
#endif

HWND mainhwnd;
static int last_keydown, last_keyup;
static int lastmx, lastmy;
static bool poll_again;
static BList queue;

static const char *key_names[] =
{
	NULL,
	"up", "down", "right", "left",
	"btn1", "btn2", "btn3", "btn4", "dpad cntr",
	
	"jogdial up", "jogdial dn", "jogdial cntr"
};

/*
0x26 0x84,	// up
0x28 0x84,	// down
0x27 0x84,	// right
0x25 0x84,	// left

0x5b 0xc1,	// btn1
0x5b 0xc2,	// btn2
0x5b 0xc3,	// btn3
0x5b 0xc4,	// btn4

0x5b 0xc5,	// dpad centered (not getting an event for this ATM...look into it)

0x26 0x83,	// jogdial up
0x28 0x83,	// jogdial up
0x86		// jogdial center
*/

bool SDLS_EventInit(void)
{
	if (create_window()) return 1;
	last_keydown = last_keyup = SDLK_UNKNOWN;
	
	return 0;
}

void SDLS_EventQuit(void)
{
	DestroyWindow(mainhwnd);
	
	while(queue.CountItems())
		delete (SDL_Event *)queue.RemoveItem(queue.CountItems() - 1);
}

/*
void c------------------------------() {}
*/

int SDL_PollEvent(SDL_Event *event)
{
	SDL_PumpEvents();
	
	if (event)
	{
		SDL_Event *nextevt = (SDL_Event *)queue.RemoveItem(0);
		if (nextevt)
		{
			*event = *nextevt;
			delete nextevt;
			return 1;
		}
	}
	
	return 0;
}


void SDL_PumpEvents()
{
MSG msg;
int count = 10;

	do
	{
		if (!PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) break;
		
		poll_again = false;
		
		TranslateMessage(&msg);
		DispatchMessage(&msg);
		
		if (!poll_again) count--;
	}
	while(count);
}


LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
SDL_Event event;

    switch(message)
    {
		case WM_KEYDOWN:
		{
			int sdlk = vk_to_sdlk(wParam, last_keydown, &poll_again);
			last_keydown = wParam;
			
			if (sdlk != SDLK_UNKNOWN)
			{
				event.type = SDL_KEYDOWN;
				event.key.keysym.sym = sdlk;
				event.key.keysym.scancode = 0;
				
				SDL_PushEvent(&event);
			}
		}
		break;
		
		case WM_KEYUP:
		{
			int sdlk = vk_to_sdlk(wParam, last_keydown, &poll_again);
			last_keydown = wParam;
			
			if (sdlk != SDLK_UNKNOWN)
			{
				event.type = SDL_KEYUP;
				event.key.keysym.sym = sdlk;
				event.key.keysym.scancode = 0;
				
				SDL_PushEvent(&event);
			}
		}
		break;
		
		case WM_LBUTTONDOWN:
		{
			event.type = SDL_MOUSEBUTTONDOWN;
			event.button.x = lastmx = GET_X_LPARAM(lParam);
			event.button.y = lastmy = GET_Y_LPARAM(lParam);
			event.button.button = 1;
			
			event.type = SDL_QUIT;	// hack
			SDL_PushEvent(&event);
		}
		break;
		
		case WM_LBUTTONUP:
		{
			event.type = SDL_MOUSEBUTTONUP;
			event.button.x = GET_X_LPARAM(lParam);
			event.button.y = GET_Y_LPARAM(lParam);
			event.button.button = 1;
			
			SDL_PushEvent(&event);
		}
		break;
		
		case WM_MOUSEMOVE:
		{
			event.type = SDL_MOUSEMOTION;
			event.motion.x = GET_X_LPARAM(lParam);
			event.motion.y = GET_Y_LPARAM(lParam);
			event.motion.xrel = (event.motion.x - lastmx);
			event.motion.yrel = (event.motion.y - lastmy);
			
			lastmx = event.motion.x;
			lastmy = event.motion.y;
			
			SDL_PushEvent(&event);
		}
		break;
	}
	
	return 0;
}


void SDL_PushEvent(SDL_Event *event)
{
	SDL_Event *evt = new SDL_Event;
	*evt = *event;
	
	queue.AddItem(evt);
}


const char *SDL_GetKeyName(SDLKey key)
{
	if (key >= 0 && key < SDLK_LAST)
		return key_names[key];
	
	return NULL;
}


void SDL_WarpMouse(uint16_t x, uint16_t y)
{
SDL_Event event;

	event.type = SDL_MOUSEMOTION;
	event.motion.x = x;
	event.motion.y = y;
	event.motion.xrel = (x - lastmx);
	event.motion.yrel = (y - lastmy);
	lastmx = x;
	lastmy = y;
	
	SDL_PushEvent(&event);
}


void SDL_WM_SetCaption(const char *title, const char *icon)
{
	WCHAR *wtitle = (WCHAR *)malloc(strlen(title) * 2);
	for(int i=0;;i++)
	{
		wtitle[i] = title[i];
		if (!title[i]) break;
	}
	
	SetWindowText(mainhwnd, wtitle);
}


uint8_t SDL_GetAppState(void)
{
	return (SDL_APPACTIVE | SDL_APPINPUTFOCUS | SDL_APPMOUSEFOCUS);
}

/*
void c------------------------------() {}
*/

static bool create_window(void)
{
	const WCHAR *classname = L"SDL Window";
	WNDCLASS wcex;
	
    wcex.style          = CS_HREDRAW | CS_VREDRAW;
    wcex.lpfnWndProc    = WndProc;
    wcex.cbClsExtra     = 0;
    wcex.cbWndExtra     = 0;
    wcex.hInstance      = GetModuleHandle(NULL);
    wcex.hIcon          = NULL;
    wcex.hCursor        = LoadCursor(NULL, IDC_ARROW);
    wcex.hbrBackground  = (HBRUSH)(COLOR_WINDOW+1);
    wcex.lpszMenuName   = NULL;
    wcex.lpszClassName  = classname;
	
	if (!RegisterClass(&wcex))
	{
		MSGBOX("Window Class Register failed!");
		return 1;
	}
	
	mainhwnd = CreateWindow(classname, L"SDL Window", WS_OVERLAPPEDWINDOW, \
    					0, 0, 240, 320, NULL, NULL, wcex.hInstance, NULL);
	if (!mainhwnd)
	{
		MSGBOX("Create Main Window failed!");
		return 1;
	}
	
	ShowWindow(mainhwnd, SW_SHOW);
	return 0;
}



static int vk_to_sdlk(uint8_t vcode, uint8_t last, bool *is_seq_start)
{
	switch(vcode)
	{
		case 0x25:
		case 0x26:
		case 0x27:
		case 0x28:
		case 0x5b:
			*is_seq_start = true;
		break;
		
		case 0x84:		// arrows
		{
			switch(last)
			{
				case 0x26: return SDLK_UP;
				case 0x28: return SDLK_DOWN;
				case 0x27: return SDLK_RIGHT;
				case 0x25: return SDLK_LEFT;
			}
		}
		break;
		
		case 0xC1: if (last == 0x5b) return SDLK_BTN1;
		case 0xC2: if (last == 0x5b) return SDLK_BTN2;
		case 0xC3: if (last == 0x5b) return SDLK_BTN3;
		case 0xC4: if (last == 0x5b) return SDLK_BTN4;
		
		case 0x83:		// jog dial
		{
			switch(last)
			{
				case 0x26: return SDLK_JOGDIAL_UP;
				case 0x28: return SDLK_JOGDIAL_DOWN;
			}
		}
		break;
		
		case 0x86: return SDLK_JOGDIAL_CENTER;
	}
	
	return SDLK_UNKNOWN;
}





